# Copyright (C) 2017-2023 J.F.Dockes
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Translate an UPnP search string into a Recoll one and run the search. Note that the translation 
is not exact, we are just making a best effort."""

import sys
import re

from recoll import recoll

from upmplgutils import uplog
from conftree import stringToStrings
import uprclutils
import uprclinit


def _getchar(s, i):
    if i < len(s):
        return i + 1, s[i]
    else:
        return i, None


def _readword(s, i):
    """Extract word from whole string s starting at offset i. Return final position and word.
    The final position will point to whitespace or EOS"""
    # uplog(f"_readword: input: <{s[i:]}>")
    w = ""
    for j in range(i, len(s)):
        if s[j].isspace():
            break
        w += s[j]
    # If we broke at the end of string, increment to indicate eos
    if j == len(s) - 1:
        j += 1
    # uplog(f"_readword returning index {j} word <{w}>")
    return j, w


def _parsestring(s, i):
    """Parse double-quoted string looking like ["hello \"one phrase\" world"]
    Called with the first quote already read.
    Upnp search term strings are double quoted, but we should not take them as recoll phrases.
    We separate parts which are internally (backslash-escaped) quoted, and become phrases, and
    lists of words which we interpret as an AND search (comma-separated).
    Note that we don't handle possible quotes inside an escape-quoted phrase string."""

    # uplog(f"parseString: input: <{s[i:]}>")
    str = ""
    escape = False
    instring = False
    for j in range(i, len(s)):
        if instring:
            if escape:
                if s[j] == '"':
                    str += '"'
                    instring = False
                else:
                    str += "\\" + s[j]
                escape = False
            else:
                if s[j] == "\\":
                    escape = True
                else:
                    str += s[j]
        else:  # Not inside internal string
            if escape:
                str += s[j]
                escape = False
                if s[j] == '"':
                    instring = True
            else:
                if s[j] == "\\":
                    escape = True
                elif s[j] == '"':
                    j += 1
                    break
                else:
                    str += s[j]

    tokens = stringToStrings(str)
    # uplog(f"parseString: return: j {j} tokens {tokens}")
    return j, tokens


def _separatePhrasesAndWords(v):
    """Built separate lists for phrases and words from parseString output"""
    swords = ""
    phrases = []
    for w in v:
        if len(w.split()) == 1:
            if swords:
                swords += ","
            swords += w
        else:
            phrases.append(w)
    return (swords, phrases)


def _searchClauses(out, neg, field, oper, words, phrases):
    """Build recoll search clauses out of list of words and phrases and given negation, field name
    and operator, e.g. [title:word, title:"some phrase"]"""
    if words:
        if neg:
            out.append("-")
        out.append(field)
        out.append(oper)
        out.append(words)
    for ph in phrases:
        if neg:
            out.append("-")
        out.append(field)
        out.append(oper)
        out.append('"' + ph + '"')
    return out


def _makeSearchExp(out, v, fields, oper, neg):
    """Process data from an UPnP relExp triplet.
    Build a recoll search expression, given:
    v: terms or phrases out of _parseString.
    fields: one or several fields to be ORed (e.g. title -> title or filename)
    oper: :, = etc. "I" for ignore.
    neg: exclusion"""

    # uplog(f"_makeSearchExp: v <{v}> field <{field}> oper <{oper}> neg <{neg}>")

    if oper == "I":
        return

    # Test coming from, e.g. <upnp:class derivedfrom/= object.container.album>
    if (oper == ":" or oper == "=") and len(v) == 1:
        if v[0].startswith("object.container"):
            v = [
                "inode/directory",
            ]
        elif v[0].startswith("object.item"):
            neg = True
            v = [
                "inode/directory",
            ]

    swords, phrases = _separatePhrasesAndWords(v)

    # Special-case 'title' because we want to also match the file name. Possibly only for
    # directories, but there is no way to do inside a single query (see comment further).
    if isinstance(fields, str):
        fields = (fields,)

    if len(fields) > 1:
        out.append(" (")

    for i in range(len(fields)):
        field = fields[i]
        out.append(" (")
        _searchClauses(out, neg, field, oper, swords, phrases)
        # We'd like to do the following to avoid matching regular file names (maybe: what of the
        # untagged files?) but recoll takes all mime: clause as global filters, so can't work
        # anyway:
        # if i == 1: out.append(" AND mime:inode/directory")
        out.append(")")
        if len(fields) > 1 and i < len(fields) -1:
            if neg:
                out.append(" AND ")
            else:
                out.append(" OR ")

    if len(fields) > 1:
        out.append(") ")


# Some fields are found in searches but not in the upnp2rclfield dict, which would lead to a search
# failure
_fieldaliases = {
    "dc:creator": "upnp:artist",
}
# Some upnp fields need to be expanded to multiple recoll ones and result in an OR search
# Note that upnp has no notion at all of "composer" or "conductor", only upnp:artist or dc:creator.
# The "role" value is free text as far as I can see.
_fieldexpansions = {
    "upnp:artist": ("artist", "composer", "conductor", "author"),
    "dc:title": ("title", "filename"),
}

# Upnp searches are made of relExps which are always (selector, operator, value), there are no unary
# operators. The relExp-s can be joined with and/or and grouped with parentheses, but they are
# always triplets, which make things reasonably easy to translate into a recoll search, just
# translating the relExps into field clauses and forwarding the booleans and parentheses.
#
# Also the set of unquoted keywords or operators is unambiguous and all un-reserved values are
# quoted, so that we don't even use a state machine, but rely of comparing token values for
# determining our syntactic state.
#
# This is all quite approximative though, but simpler than a formal parser, and mostly works in
# practise because we rely on recoll to deal with logical operators and parentheses.
# UPnP search string grammar as from the standard:
#    searchCrit ::= searchExp | asterisk
#    searchExp  ::= relExp |
#                   searchExp wChar+ logOp wChar+ searchExp |
#                   '(' wChar* searchExp wChar* ')'
#    logOp      ::= 'and' | 'or'
#    relExp     ::= property wChar+ binOp wChar+ quotedVal |
#                   property wChar+ existsOp wChar+ boolVal
#    binOp      ::= relOp |stringOp
#    relOp      ::= '=' | '!=' | '<' | '<=' | '>' | '>='
#    stringOp   ::= 'contains' | 'doesNotContain' | 'derivedFrom'
#    existsOp   ::= 'exists'
#    boolVal    ::= 'true' | 'false'
#    wChar      ::= space | hTab | lineFeed | vTab | formFeed | return
#    property   ::= (* property name as defined in section 2.4 *)
#    escapedQuote ::= (* double-quote escaped string as defined in section 2.3.1 *)
def _upnpsearchtorecoll(s):
    uplog(f"_upnpsearchtorecoll:in: <{s}>")

    # Simplify: turn all whitespace sequences into single spaces
    s = re.sub(r"[\t\n\r\f ]+", " ", s)

    out = []
    field = ""
    oper = ""
    neg = False
    i = 0
    while True:
        i, c = _getchar(s, i)
        if not c:
            break
        # uplog(f"_upnpsearchtorecoll: nextchar: <{c}>")

        if c.isspace():
            continue

        if c == "*":
            if (len(out) > 1 or (len(out) == 1 and not out[-1].isspace())) or (
                len(s[i:]) and not s[i:].isspace()
            ):
                raise Exception("If * is used it must be the only input")
            out = ["mime:*"]
            break

        if c == "(" or c == ")":
            # uplog(f"Appending <{c}>")
            out.append(c)
        elif c == ">" or c == "<" or c == "=":
            oper += c
        else:
            if c == '"':
                # Quoted strings are always values (3rd triplet elements) and values are always
                # quoted. Read it, then generate and append one or several recoll search clauses,
                # according to the already read 1st and 2nd triplet elements.

                # We must have a field and an operator at this point.
                if not field or not oper:
                    uplog(f"Search: syntax error: [{s}]")
                    return None

                i, v = _parsestring(s, i)

                # If the operator is "=" for a string field, we transform the search into an
                # anchored phrase search.
                if field != "mime" and oper == "=" :
                    fs = "^" + " ".join(v) + "$"
                    v = [fs]
                    oper = ":"

                _makeSearchExp(out, v, field, oper, neg)

                # Reset everything and go on
                field = ""
                oper = ""
                neg = False
                continue

            # Unquoted word.
            i -= 1
            i, w = _readword(s, i)

            w = w.lower()
            if w == "contains":
                oper = ":"
            elif w == "doesnotcontain":
                neg = True
                oper = ":"
            elif w == "derivedfrom":
                oper = ":"
            elif w == "exists":
                # somefield exists true
                # can't use this, will be ignored
                oper = "I"
            elif w == "true" or w == "false":
                # Don't know what to do with this. Just ignore it,
                # by not calling makeSearchExp.
                pass
            elif w == "and":
                # Recoll has implied AND, but see next
                pass
            elif w == "or":
                # Does not work because OR/AND priorities are reversed between recoll and upnp. This
                # would be very difficult to correct, let's hope that the callers use parentheses
                out.append("OR")
            else:
                if w == "upnp:class":
                    field = "mime"
                else:
                    try:
                        if w in _fieldaliases:
                            w = _fieldaliases[w]
                        if w in _fieldexpansions:
                            field = _fieldexpansions[w]
                        else:
                            field = uprclutils.upnp2rclfields[w]
                    except Exception as ex:
                        uplog(f"Field translation error for <{w}>: {ex}. Ignoring search.")
                        # Recoll would just ignore an unknown field spec and search the main
                        # text. We don't want to do this, it yields confusing results (happens for
                        # example with someting like: upnp:artist[@role="composer"] from kazoo
                        return None

    return " ".join(out)


def search(foldersobj, rclconfdir, inobjid, upnps, idprefix, httphp, pathprefix):
    """Run UPnP search operation. inobjid is for the container this search is run from."""

    tags = uprclinit.getTree("tags")

    # Translate UPnP search string to recoll one
    rcls = _upnpsearchtorecoll(upnps)

    if not rcls:
        uplog(f"Upnp search string parse failed or ignored for [{upnps}]. Recoll search is empty")
        return []
    uplog(f"Search: recoll search: <{rcls}>")

    filterdir = foldersobj.dirpath(inobjid)
    if filterdir and filterdir != "/":
        # uplog(f"filterdir: <{filterdir}>")
        rcls += ' dir:"' + filterdir + '"'

    rcldb = recoll.connect(confdir=rclconfdir)
    try:
        rclq = rcldb.query()
        rclq.execute(rcls)
    except Exception as e:
        uplog("Search: recoll query raised: %s" % e)
        return []

    uplog("Estimated query results: %d" % (rclq.rowcount))
    if rclq.rowcount == 0:
        return []

    entries = []
    maxcnt = 0
    while True:
        docs = rclq.fetchmany()
        for doc in docs:
            # The doc is either an actual recollindex product from the
            # FS or a synthetic one from uprcltags creating album
            # entries. Different processing for either
            e = None
            if doc["rcludi"].find("albid") == 0:
                albid = doc["rcludi"][5:]
                # uplog(f"Search: album: {doc['rcludi']} albid {albid}")
                e = tags.direntryforalbid(albid)
            elif doc["rcludi"].find("artid") == 0:
                artid = doc["rcludi"][5:]
                # uplog(f"Search: artist: {doc['rcludi']} albid {albid}")
                e = tags.direntryforartid(artid)
            else:
                # Objidfordoc uses the path from the url to walk the _dirvec and determine the right
                # entry if doc is a container. If doc is an item, the returned id is not usable but
                # still unique (based on the xdocid). We used to return
                # (0$uprcl$folders$seeyoulater), but this ennoys bubble to no end.  This still
                # breaks the recommendation for the objids to be consistent and unchanging
                id = foldersobj.objidfordoc(doc)
                # uplog("Search: id [%s] for doc udi [%s]\n" % (id, doc["rcludi"]))
                e = uprclutils.rcldoctoentry(id, inobjid, httphp, pathprefix, doc)
            # uplog(f"Search: entry: {e}")
            if e:
                entries.append(e)
        if (maxcnt > 0 and len(entries) >= maxcnt) or len(docs) != rclq.arraysize:
            break
    uplog("Search retrieved %d docs" % (len(entries),))

    entries.sort(key=uprclutils.cmpentries)
    return entries


if __name__ == "__main__":
    s = '(upnp:artist derivedFrom  "abc\\"def\\g") or (dc:title:xxx) '
    s = 'upnp:class derivedfrom "object.container.album" and dc:title contains "n"'
    if len(sys.argv) > 1:
        s = sys.argv[1]
    print("INPUT: %s" % s)
    o = _upnpsearchtorecoll(s)
    print("OUTPUT: %s" % o)



# Copyright (C) 2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum
from typing import Callable, List, TypeVar

from artist_metadata import ArtistMetadata
from search_util import simplify


class _SearchResultRank:
    def __init__(self, rank_value: int):
        self.__rank_value: int = rank_value

    @property
    def rank_value(self) -> int:
        return self.__rank_value


class SearchResultRank(Enum):
    MATCH_EXACT = _SearchResultRank(rank_value=1)
    MATCH_CASE_SENSITIVE = _SearchResultRank(rank_value=2)
    MATCH_CASE_INSENSITIVE = _SearchResultRank(rank_value=3)
    STARTS_WITH_EXACT = _SearchResultRank(rank_value=4)
    STARTS_WITH_CASE_SENSITIVE = _SearchResultRank(rank_value=5)
    STARTS_WITH_CASE_INSENSITIVE = _SearchResultRank(rank_value=6)
    ENDS_WITH_EXACT = _SearchResultRank(rank_value=7)
    ENDS_WITH_CASE_SENSITIVE = _SearchResultRank(rank_value=8)
    ENDS_WITH_CASE_INSENSITIVE = _SearchResultRank(rank_value=9)
    CONTAINS_EXACT = _SearchResultRank(rank_value=10)
    CONTAINS_CASE_SENSITIVE = _SearchResultRank(rank_value=11)
    CONTAINS_CASE_INSENSITIVE = _SearchResultRank(rank_value=12)

    @property
    def rank_value(self) -> int:
        return self.value.rank_value

    def is_match(self, sv: str, sv_s: str, sv_l: str, n: str, n_s: str, n_l: str) -> bool:
        """
        sv: search_value, sv_s: simplified, sv_l: lower
        n: name, n_s: name_simplified, n_l: name_lower
        """
        match self:
            case SearchResultRank.MATCH_EXACT:
                return n == sv
            case SearchResultRank.MATCH_CASE_SENSITIVE:
                return n_s == sv_s
            case SearchResultRank.MATCH_CASE_INSENSITIVE:
                return n_l == sv_l
            case SearchResultRank.STARTS_WITH_EXACT:
                return n.startswith(sv)
            case SearchResultRank.STARTS_WITH_CASE_SENSITIVE:
                return n_s.startswith(sv_s)
            case SearchResultRank.STARTS_WITH_CASE_INSENSITIVE:
                return n_l.startswith(sv_l)
            case SearchResultRank.ENDS_WITH_EXACT:
                return n.endswith(sv)
            case SearchResultRank.ENDS_WITH_CASE_SENSITIVE:
                return n_s.endswith(sv_s)
            case SearchResultRank.ENDS_WITH_CASE_INSENSITIVE:
                return n_l.endswith(sv_l)
            case SearchResultRank.CONTAINS_EXACT:
                return sv in n
            case SearchResultRank.CONTAINS_CASE_SENSITIVE:
                return sv_s in n_s
            case SearchResultRank.CONTAINS_CASE_INSENSITIVE:
                return sv_l in n_l
            case _:
                return False


def sort_artist_list_by_rank(
        search_value: str,
        artist_list: list[ArtistMetadata]) -> list[ArtistMetadata]:
    def get_names(a: ArtistMetadata) -> list[str]:
        return [a.artist_name, a.artist_sort_name] if a.artist_sort_name else [a.artist_name]
    return sort_obj_by_rank(
        search_value=search_value,
        obj_list=artist_list,
        key=get_names)


T = TypeVar("T")


def sort_obj_by_rank(
        search_value: str,
        obj_list: List[T],
        key: Callable[[T], List[str]]) -> List[T]:
    if not obj_list or len(obj_list) == 0:
        return []
    # Pre-calculate search variants once
    sv_s = simplify(search_value)
    sv_l = sv_s.lower()
    ranked_list: list[tuple[T, int]] = []
    for curr in obj_list:
        found_rank = None
        # Pre-process all names for THIS object once
        names_data = []
        for name in key(curr):
            n_s = simplify(name)
            names_data.append((name, n_s, n_s.lower()))
        # Now check those pre-processed names against each rank
        for rank_member in SearchResultRank:
            for n, n_s, n_l in names_data:
                if rank_member.is_match(search_value, sv_s, sv_l, n, n_s, n_l):
                    found_rank = rank_member.rank_value
                    break
            if found_rank:
                break
        if found_rank:
            ranked_list.append((curr, found_rank))
    if ranked_list:
        ranked_list.sort(key=lambda x: x[1])
        return [item[0] for item in ranked_list]
    return obj_list



# Copyright (C) 2023,2024,2025 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from enum import Enum
import idgenerator
import config
import constants

base_chars: str = "abcdefghijklmnopqrtuvwxyz0123456789"
current_id: int = None


class _ElementTypeData:

    def __init__(self):
        global current_id
        if current_id is None:
            current_id = 0
        else:
            current_id += 1
        self.__element_name: str = idgenerator.number_to_base_decoded(n=current_id, base_chars=base_chars)

    @property
    def element_name(self) -> str:
        return self.__element_name


class ElementType(Enum):
    TAG = _ElementTypeData()
    ALBUM = _ElementTypeData()
    ALBUM_VERSION = _ElementTypeData()
    GENRE = _ElementTypeData()
    GENRE_ARTIST_LIST = _ElementTypeData()
    GENRE_ALBUM_LIST = _ElementTypeData()
    ARTIST = _ElementTypeData()
    GENRE_ARTIST = _ElementTypeData()
    ARTIST_BY_INITIAL = _ElementTypeData()
    SONG = _ElementTypeData()
    PLAYLIST = _ElementTypeData()
    INTERNET_RADIO = _ElementTypeData()
    SONG_ENTRY_NAVIGABLE = _ElementTypeData()
    SONG_ENTRY_THE_SONG = _ElementTypeData()
    NEXT_RANDOM_SONGS = _ElementTypeData()
    NAVIGABLE_ALBUM = _ElementTypeData()
    ARTIST_TOP_SONGS = _ElementTypeData()
    ARTIST_TOP_SONGS_LIST = _ElementTypeData()
    ARTIST_SIMILAR = _ElementTypeData()
    ARTIST_ALBUMS = _ElementTypeData()
    RADIO = _ElementTypeData()
    RADIO_SONG_LIST = _ElementTypeData()
    GENRE_ARTIST_ALBUMS = _ElementTypeData()
    # artist which appear as artistId for albums
    ALBUM_FOCUS = _ElementTypeData()
    ARTIST_FOCUS = _ElementTypeData()
    ADDITIONAL_ALBUM_ARTISTS = _ElementTypeData()
    ARTIST_APPEARANCES = _ElementTypeData()
    ALBUM_SONG_SELECTION_BY_ARTIST = _ElementTypeData()
    ALBUM_DISC = _ElementTypeData()
    ARTIST_ROLE = _ElementTypeData()
    ARTIST_ROLE_INITIAL = _ElementTypeData()
    ALBUM_BROWSE_FILTER_KEY = _ElementTypeData()
    ALBUM_BROWSE_FILTER_VALUE = _ElementTypeData()
    ALBUM_BROWSE_MATCHING_ALBUMS = _ElementTypeData()
    FAVORITE_SONGS_AS_CONTAINERS = _ElementTypeData()
    FAVORITE_SONGS_CONTAINER = _ElementTypeData()

    @property
    def element_name(self) -> str:
        if config.get_config_param_as_bool(constants.ConfigParam.MINIMIZE_IDENTIFIER_LENGTH):
            return self.name
        else:
            return self.value.element_name


def get_element_type_by_name(element_name: str) -> ElementType:
    element: ElementType
    for element in ElementType:
        if element.element_name == element_name:
            return element
    raise Exception(f"get_element_type_by_name with {element_name} NOT found")



# Copyright (C) 2017 J.F.Dockes
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

import conftree
import copy
import locale
import os
import shutil
import subprocess
import sys
import time

from upmplgutils import uplog, getOptionValue
import uprclinit


# We recreate the Recoll configuration files at each startup. This is fast and avoids having to test
# for changed parameters. No directly user-entered data lives in there, it all comes from
# upmpdcli.conf, uprclconfrecolluser or the Minim configuration.
def _initconfdir(confdir, topdirs):
    if not os.path.isdir(confdir):
        if os.path.exists(confdir):
            raise Exception("Exists and not directory: %s" % confdir)
        os.mkdir(confdir)

    datadir = os.path.dirname(__file__)
    uplog("datadir: %s" % datadir)

    for fn in ("fields", "mimemap", "mimeconf"):
        dst = os.path.join(confdir, fn)
        src = os.path.join(datadir, "rclconfig-" + fn)
        shutil.copyfile(src, dst)
    try:
        with open(os.path.join(confdir, "backends"), "wb") as f:
            pass
    except Exception as err:
        uplog(f"Failed truncating backends: {err}")

    # Update "stored" section of the "fields" file according to minim config: some custom
    # tags may nnot be in our predefined list
    fields = uprclinit.allMinimTags()
    dst = os.path.join(confdir, "fields")
    fldconf = conftree.ConfSimple(dst, False, False)
    for fld in fields:
        fldconf.set(fld, "", "stored")

    exclpats = uprclinit.g_minimconfig.getexcludepatterns()

    userconfig = getOptionValue("uprclconfrecolluser")
    if not userconfig:
        userconfig = os.path.join(confdir, "recoll.conf.user")
    if os.path.exists(userconfig):
        userconfdata = open(userconfig, "rb").read()
    else:
        userconfdata = b""

    path = os.path.join(confdir, "recoll.conf")
    f = open(path, "wb")
    f.write(b"topdirs = %s\n" % topdirs.encode(locale.getpreferredencoding()))
    f.write(b"idxabsmlen = 0\n")
    f.write(b"loglevel = 2\n")
    # f.write(b"idxlogfilename = /tmp/loguprcl.txt\n")
    # f.write(b"pylogfilename = /tmp/logpyuprcl.txt\n")
    f.write(b"noaspell = 1\n")
    f.write(b"nomd5types = rclaudio rclaudio.py rclimg rclimg.py\n")
    f.write(b"testmodifusemtime = 1\n")
    f.write(b"idxmetastoredlen = 20000\n")
    # We'd like to set this to zero, but this triggers the creation of
    # a chert index through a stub file, and, for some reason, this
    # does not seem to work any more with Debian Xapian 1.4.3, I get a
    # chert backend disabled, even if XAPIAN_HAS_CHERT_BACKEND is set
    # in version.h
    f.write(b"indexStoreDocText = 1\n")
    f.write(
        b"audiotagfixerscript = %b\n" % os.path.join(datadir, "minimtagfixer.py").encode("utf-8")
    )
    f.write(b"indexallfilenames = 0\n")
    f.write(
        b"noContentSuffixes+ = .tiff .tif .xcf .bmp .xpm .svg .djvu\n"
    )
    if exclpats:
        f.write(b"skippedNames+ = " + exclpats.encode("utf-8") + b"\n")
    else:
        f.write(b"skippedNames+ = \n")
    if userconfdata:
        f.write(userconfdata)
    f.close()


_idxproc = None
_lastidxstatus = None


def runindexer(confdir, topdirs, rebuild=False):
    global _idxproc, _lastidxstatus
    if _idxproc is not None:
        raise Exception("uprclrunindexer: already running")

    _initconfdir(confdir, topdirs)

    cf = conftree.ConfSimple(os.path.join(confdir, "recoll.conf"), readonly=False)
    td = cf.get("topdirs", "")
    if td != topdirs:
        cf.set("topdirs", topdirs)

    env = copy.deepcopy(os.environ)
    env["HOME"] = confdir
    if rebuild:
        _idxproc = subprocess.Popen(["recollindex", "-c", confdir, "-z"])
    else:
        _idxproc = subprocess.Popen(["recollindex", "-c", confdir])


def indexerdone():
    global _idxproc, _lastidxstatus
    if _idxproc is None:
        return True
    _lastidxstatus = _idxproc.poll()
    if _lastidxstatus is None:
        return False
    _idxproc = None
    return True


def indexerstatus():
    return _lastidxstatus


# Only used for testing
if __name__ == "__main__":
    if len(sys.argv) != 3:
        print("Usage: uprclindex.py <confdir> <topdirs>", file=sys.stderr)
        sys.exit(1)
    runindexer(sys.argv[1], sys.argv[2])
    while True:
        if indexerdone():
            uplog("Indexing done, status: %d" % indexerstatus())
            sys.exit(0)
        uplog("Waiting for indexer")
        time.sleep(1)



# Copyright (C) 2023 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum


class _OptionKeyData:

    def __init__(self, name: str, default_value: any = None):
        self.__name: str = name
        self.__default_value: any = default_value

    @property
    def name(self) -> str:
        return self.__name

    @property
    def default_value(self) -> any:
        return self.__default_value


class OptionKey(Enum):
    SKIP_TRACK_NUMBER = _OptionKeyData("skip-track-number", False)
    FORCED_TRACK_NUMBER = _OptionKeyData("forced-track-number", None)
    SKIP_TRACK_ARTIST = _OptionKeyData("skip-track-artist", False)
    OVERRIDDEN_ART_URI = _OptionKeyData("overriden-art-uri", None)
    OVERRIDDEN_TRACK_NAME = _OptionKeyData("overridden-track-name", None)
    ENTRY_AS_CONTAINER = _OptionKeyData("entry-as-container", False)
    ADD_EXPLICIT = _OptionKeyData("add-explicit", True)
    ADD_ALBUM_YEAR = _OptionKeyData("add-album-year", True)
    SKIP_ART = _OptionKeyData("skip-art", False)
    ADD_ARTIST_TO_ALBUM_ENTRY = _OptionKeyData("add-artist-to-album-entry", False)
    # integer to prepend e.g. pass 3 -> [03] album_title instead of album_title
    PREPEND_ENTRY_NUMBER_IN_ENTRY_NAME = _OptionKeyData("prepend-entry-number-in-entry-name", None)
    OMIT_ARTIST_TO_ALBUM_ENTRY_UNLESS_DIFFERENT = _OptionKeyData("omit-rsts-nls-dffrnt", False)
    ALBUM_OMITTABLE_ARTIST_ID = _OptionKeyData("lbm-omit-rtst", None)
    TRACK_OMITTABLE_ARTIST_NAME = _OptionKeyData("trk-omit-rtst", None)
    INITIAL_TRACK_NUMBER = _OptionKeyData("ntl-trkn", 1)
    TRACK_AS_NAVIGABLE = _OptionKeyData("nvgbl-trk", False)
    TRACK_CONTAINER_SET_CLASS = _OptionKeyData("trkc-sclzz", False)

    @property
    def name(self) -> str:
        return self.value.name

    @property
    def default_value(self) -> any:
        return self.value.default_value



# Copyright (C) 2026 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class DiscTitle:

    def __init__(self, disc_num: int, title: str):
        self.__disc_num: str = disc_num
        self.__title: str = title

    @property
    def disc_num(self) -> str:
        return self.__disc_num

    @property
    def title(self) -> str:
        return self.__title



# Copyright (C) 2023 Giovanni Fulco
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum


class ItemIdentifierKey(Enum):
    THING_NAME = 0, 'n'
    THING_VALUE = 1, 'v'
    PAGE_NUMBER = 3, 'p'
    ALBUM_ID = 4, 'a'
    OFFSET = 5, 'o'
    TAG_TYPE = 6, 't',
    ALBUM_PAGE = 7, 'ap'
    CATEGORY_TITLE = 10, 'catt'
    PAGE_LIST = 11, 'pglst'
    PAGE_LINK_API_PATH = 12, "pglnkppth"
    CATEGORY_KEY = 13, "catk"
    MAX_ITEMS = 14, "mxtms"
    MISSING_ARTIST_ART = 15, "msngrtrt"
    COUNT = 16, "cnt"
    UNDERLYING_TYPE = 17, "ut"
    LISTEN_QUEUE_ACTION = 18, "lqctn"
    LAST_FOUND_ID = 19, "lfid"
    FAVORITE_ACTION = 20, "fvctn"
    RANDOM_VALUE = 21, "rnd"
    SEARCH_FIELD = 22, "srchfld"
    SEARCH_KIND = 23, "srchknd"

    def __init__(
            self,
            num: int,
            key_name: str):
        self.num: int = num
        self.key_name: str = key_name

    def getName(self) -> str:
        return self.key_name


# duplicate check
name_checker_set: set[str] = set()
id_checker_set: set[int] = set()
for v in ItemIdentifierKey:
    if v.getName() in name_checker_set:
        raise Exception(f"Duplicated name [{v.getName()}]")
    if v.value[0] in id_checker_set:
        raise Exception(f"Duplicated id [{v.value[0]}]")
    name_checker_set.add(v.getName())
    id_checker_set.add(v.value[0])
